/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/* Multi-user networking protocol implementation for TCP/IP on BSD UNIX */

#include "my-inet.h"		/* inet_addr() */
#include <errno.h>		/* EMFILE, EADDRNOTAVAIL, ECONNREFUSED,
				 * ENETUNREACH, ETIMEOUT */
#include "my-in.h"		/* struct sockaddr_in, INADDR_ANY, htons(),
				 * htonl(), ntohl(), struct in_addr */
#include <netdb.h>		/* struct hostent, gethostbyaddr() */
#include "my-socket.h"		/* socket(), AF_INET, SOCK_STREAM,
				 * setsockopt(), SOL_SOCKET, SO_REUSEADDR,
				 * bind(), struct sockaddr, accept(),
				 * shutdown(), connect() */
#include "my-string.h"		/* strtol(), memcpy() */
#include "my-unistd.h"		/* close() */

#include "config.h"
#include "exceptions.h"
#include "log.h"
#include "options.h"
#include "timers.h"
#include "net_proto.h"

#define DEFAULT_PORT 7777

static int server_socket;

const char *
proto_name(void)
{
    return "BSD/TCP";
}

const char *
proto_usage_string(void)
{
    return "[port]";
}

int
proto_initialize(struct proto *proto, int argc, char **argv)
{
    int			port;
    char	       *p;
    struct sockaddr_in	address;
    int			option = 1;
    
    proto->pocket_size = 1;
    proto->believe_eof = 1;
    proto->eol_out_string = "\r\n";
    proto->eol_in_char = '\n';

    if (argc > 1)
	return 0;
    else if (argc == 1) {
	port = strtol(argv[0], &p, 10);
	if (*p != '\0')
	    return 0;
    } else
	port = DEFAULT_PORT;

    server_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (server_socket < 0) {
	log_perror("Creating server socket");
	return 0;
    }
    if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR,
		   (char *) &option, sizeof(option)) < 0) {
	log_perror("Setting server socket options");
	return 0;
    }
    address.sin_family = AF_INET;
    address.sin_addr.s_addr = htonl(INADDR_ANY);
    address.sin_port = htons(port);
    if (bind(server_socket, (struct sockaddr *) &address, sizeof(address))) {
	log_perror("Binding server socket");
	close(server_socket);
	return 0;
    }
    proto->connection_fd = server_socket;
    return 1;
}

void
proto_listen(struct proto *proto)
{
    listen(server_socket, 5);
}

static Exception	ns_timeout;

static void
name_server_timeout(Timer_ID id, Timer_Data data)
{
    RAISE(ns_timeout, 0);
}

static const char *
address_string(long a)
{
    static char buf[20];

    sprintf(buf, "%d.%d.%d.%d", (a >> 24) & 0xff, (a >> 16) & 0xff,
	    (a >> 8) & 0xff, a & 0xff);
    return buf;
}

enum proto_accept_error
proto_accept_connection(int *read_fd, int *write_fd, const char **name)
{
    volatile int	fd;
    struct sockaddr_in	address;
    int			addr_length = sizeof(address);
    struct hostent     *e = 0;
    Timer_ID		id;
    const char	       *host_name;

    fd = accept(server_socket, (struct sockaddr *) &address, &addr_length);
    if (fd < 0) {
	if (errno == EMFILE)
	    return PA_FULL;
	else {
	    log_perror("Accepting new network connection");
	    return PA_OTHER;
	}
    }
    id = set_timer(5, name_server_timeout, 0);
    TRY
	e = gethostbyaddr((void *) &address.sin_addr,
			  sizeof(address.sin_addr),
			  AF_INET);
	cancel_timer(id);
    EXCEPT (ns_timeout)
	e = 0;
	reenable_timers();
    ENDTRY;
    host_name = e ? e->h_name : address_string(ntohl(address.sin_addr.s_addr));

    *read_fd = *write_fd = fd;
    *name = host_name;
    return PA_OKAY;
}

void
proto_close_connection(int read_fd, int write_fd)
{
    /* read_fd and write_fd are the same, so we only need to deal with one. */
    shutdown(read_fd, 2);
    close(read_fd);
}

void
proto_shutdown(void)
{
    close(server_socket);
}

#ifdef OUTBOUND_NETWORK
#include "structures.h"

enum error
proto_open_connection(Var arglist, int *read_fd, int *write_fd)
{
    /* These are `static' rather than `volatile' because I can't cope with
     * getting all those nasty little parameter-passing rules right.  This
     * function isn't recursive anyway, so it doesn't matter.
     */
    static const char		       *host_name;
    static int				port;
    static struct hostent	       *e;
    static struct sockaddr_in		addr;
    Timer_ID				id;
    int					s;

    if (arglist.v.list[0].v.num != 2)
	return E_ARGS;
    else if (arglist.v.list[1].type != TYPE_STR ||
	     arglist.v.list[2].type != TYPE_NUM)
	return E_TYPE;

    host_name = arglist.v.list[1].v.str;
    port = arglist.v.list[2].v.num;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    id = set_timer(5, name_server_timeout, 0);
    TRY
	e = gethostbyname(host_name);
	cancel_timer(id);
    EXCEPT (ns_timeout)
	e = 0;
	reenable_timers();
    ENDTRY;
    if (e)
	memcpy(&addr.sin_addr, e->h_addr_list[0], e->h_length);
    else {
	struct in_addr	ia;

	ia.s_addr = inet_addr(host_name);
	if (ia.s_addr == -1)
	    return E_INVARG;
	memcpy(&addr.sin_addr, &ia, sizeof(ia));
    }

    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	if (errno != EMFILE)
	    log_perror("Making socket in proto_open_connection");
	return E_QUOTA;
    }
    if (connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
	close(s);
	if (errno == EADDRNOTAVAIL ||
	    errno == ECONNREFUSED  ||
	    errno == ENETUNREACH   ||
	    errno == ETIMEDOUT)
	    return E_INVARG;
	log_perror("Connecting in proto_open_connection");
	return E_QUOTA;
    }

    *read_fd = *write_fd = s;

    return E_NONE;
}
#endif /* OUTBOUND_NETWORK */

char rcsid_net_bsd_tcp[] = "$Id: net_bsd_tcp.c,v 1.7 1992/10/23 23:03:47 pavel Exp $";

/* $Log: net_bsd_tcp.c,v $
 * Revision 1.7  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.6  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.5  1992/10/17  20:43:13  pavel
 * Changed to use "my-inet.h" rather than <arpa/inet.h>.
 *
 * Revision 1.4  1992/10/06  01:36:45  pavel
 * Moved non-blocking code to net_multi.c, replacing it with code to set
 * proto->believe_eof appropriately.
 *
 * Revision 1.3  1992/10/04  04:58:03  pavel
 * Fixed #include's for declaration of inet_addr().  Also replaced INADDR_ANY
 * by htonl(INADDR_ANY), for portability.
 *
 * Revision 1.2  1992/09/26  02:22:53  pavel
 * Added support for printing the network protocol name on server start-up.
 *
 * Revision 1.1  1992/09/23  17:14:17  pavel
 * Initial RCS-controlled version.
 *
 */
