/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-types.h"	/* must be first on some systems */
#include "my-signal.h"
#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-unistd.h"
#include "my-wait.h"

#include "config.h"
#include "db.h"
#include "functions.h"
#include "list.h"
#include "log.h"
#include "network.h"
#include "options.h"
#include "parse_cmd.h"
#include "random.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "tasks.h"
#include "timers.h"
#include "utils.h"
#include "version.h"

#include "execute.h"

static pid_t		parent_pid;
int			in_child = 0;

static const char      *shutdown_message = 0;	/* shut down if non-zero */

typedef enum {CHKPT_OFF, CHKPT_TIMER, CHKPT_SIGNAL} Checkpoint_Reason;
static Checkpoint_Reason checkpoint_requested = CHKPT_OFF;

typedef struct shandle {
    struct shandle     *next, **prev;
    network_handle	nhandle;
    time_t		connection_time;
    time_t		last_activity_time;
    Objid		player;
    task_queue		tasks;
    int			disconnect_me;
} shandle;

shandle	*active_shandles = 0;


static void
free_shandle(shandle *h)
{
    *(h->prev) = h->next;
    if (h->next)
	h->next->prev = h->prev;

    free_task_queue(h->tasks);

    myfree(h, M_NETWORK);
}

static void
send_shutdown_message(const char *msg)
{
    shandle	*h;
    Stream	*s = new_stream(100);
    char	*message;

    stream_printf(s, "*** Shutting down: %s ***", msg);
    message = stream_contents(s);
    for (h = active_shandles; h; h = h->next)
	network_send_line(h->nhandle, message);
    free_stream(s);
}

static void
abort_server(void)
{
    signal(SIGINT,  SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    signal(SIGFPE,  SIG_DFL);
    signal(SIGHUP,  SIG_DFL);
    signal(SIGILL,  SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGSEGV, SIG_DFL);
    signal(SIGBUS,  SIG_DFL);
    signal(SIGUSR1, SIG_DFL);
    signal(SIGUSR2, SIG_DFL);
    signal(SIGCHLD, SIG_DFL);

    abort();
}

void
panic(const char *message)
{
    static int	in_panic = 0;

    errlog("PANIC%s: %s\n", in_child ? " (in child)" : "", message);
    if (in_panic) {
	errlog("RECURSIVE PANIC: aborting\n");
	abort_server();
    }
    in_panic = 1;

    log_command_history();

    if (in_child) {		/* We're a forked checkpointer */
	errlog("Child shutting down parent via USR1 signal\n");
	kill(parent_pid, SIGUSR1);
	_exit(1);
    }

    send_shutdown_message("server panic");
    network_shutdown();
    db_flush(FLUSH_PANIC);

    abort_server();
}

static void
panic_signal(int sig)
{
    char message[100];
    
    sprintf(message, "Caught signal %d", sig);
    panic(message);
}

static void
shutdown_signal(int sig)
{
    shutdown_message = "shutdown signal received";
}

static void
checkpoint_signal(int sig)
{
    checkpoint_requested = CHKPT_SIGNAL;

    signal(sig, checkpoint_signal);
}

static void
child_completed_signal(int sig)
{
#if HAVE_WAITPID
    while (waitpid(-1, 0, WNOHANG) > 0)
	;
#else
#if HAVE_WAIT3
    while (wait3(0, WNOHANG, 0) >= 0)
	;
#else
#if HAVE_WAIT2
    while (wait2(0, WNOHANG) >= 0)
	;
#else
    #error I need a non-blocking wait()-like primitive!
#endif
#endif
#endif

    signal(sig, child_completed_signal);
}

static void
setup_signals(void)
{
    signal(SIGFPE,  panic_signal);
    signal(SIGHUP,  panic_signal);
    signal(SIGILL,  panic_signal);
    signal(SIGQUIT, panic_signal);
    signal(SIGSEGV, panic_signal);
    signal(SIGBUS,  panic_signal);

    signal(SIGINT,  shutdown_signal);
    signal(SIGTERM, shutdown_signal);
    signal(SIGUSR1, shutdown_signal);   /* remote shutdown signal */
    signal(SIGUSR2, checkpoint_signal); /* remote checkpoint signal */

    signal(SIGCHLD, child_completed_signal);
}

static void
checkpoint_timer(Timer_ID id, Timer_Data data)
{
    checkpoint_requested = CHKPT_TIMER;
}

static void
set_checkpoint_timer(int first_time)
{
    Var			v;
    int			interval;
    static Timer_ID	last_checkpoint_timer;

    v = get_system_property("dump_interval");
    if (v.type != TYPE_NUM  ||  v.v.num < 60)
	interval = 3600;	/* Once per hour */
    else
	interval = v.v.num;

    if (!first_time)
	cancel_timer(last_checkpoint_timer);
    last_checkpoint_timer = set_timer(interval, checkpoint_timer, 0);
}

static const char *
object_name(Objid oid)
{
    static Stream      *s = 0;

    if (!s)
	s = new_stream(30);

    if (valid(oid))
	stream_printf(s, "%s (#%d)", obj_name(oid), oid);
    else
	stream_printf(s, "#%d", oid);

    return reset_stream(s);
}

static void
run_confuncs(Objid player)
{
    if (valid(player)) {
	Objid	location = obj_location(player);

	if (location != NOTHING) {
	    Var	args;

	    args = new_list(1);
	    args.v.list[1].type = TYPE_OBJ;
	    args.v.list[1].v.obj = player;
	    run_server_task(player, location, "confunc", args, 0);
	}
	run_server_task(player, player, "confunc", new_list(0), 0);
    }
}

static void
run_disfuncs(Objid player)
{
    if (valid(player)) {
	Objid	location = obj_location(player);

	run_server_task(player, player, "disfunc", new_list(0), 0);

	if (location != NOTHING) {
	    Var	args;

	    args = new_list(1);
	    args.v.list[1].type = TYPE_OBJ;
	    args.v.list[1].v.obj = player;
	    run_server_task(player, location, "disfunc", args, 0);
	}
    }
}

static void
main_loop(void)
{
    run_server_task(-1, SYSTEM_OBJECT, "server_started", new_list(0), 0);
    set_checkpoint_timer(1);

    while (shutdown_message == 0) {
	/* Check how long we have until the next task will be ready to run.
	 * We only care about three cases (== 0, == 1, and > 1), so we can
	 * map a `never' result from the task subsystem into 2.
	 */
	int		task_seconds = next_task_start();
	int		seconds_left = task_seconds < 0 ? 2 : task_seconds;
	shandle	       *h, *nexth;
	
	if (checkpoint_requested != CHKPT_OFF) {
	    if (checkpoint_requested == CHKPT_SIGNAL)
		log("CHECKPOINTING due to remote request signal.\n");
	    checkpoint_requested = CHKPT_OFF;
	    db_flush(FLUSH_ALL_NOW);
	    set_checkpoint_timer(0);
	}
	    
	if (!network_process_io(seconds_left ? 1 : 0)  &&  seconds_left > 1)
	    db_flush(FLUSH_ONE_SECOND);
	else
	    db_flush(FLUSH_IF_FULL);
	
	run_ready_tasks();

	{ /* Get rid of old un-logged-in or useless connections */
#ifdef CONNECT_TIMEOUT
	    int	now = time(0);
#endif
	    for (h = active_shandles; h; h = nexth) {
		nexth = h->next;
#ifdef CONNECT_TIMEOUT
		if (h->connection_time == 0
		    &&  now - h->last_activity_time > CONNECT_TIMEOUT) {
		    log("TIMEOUT: #%d on %s\n",
			h->player,
			network_connection_name(h->nhandle));
		    network_send_line(h->nhandle,
				      "*** Timed-out waiting for login. ***");
		    network_close(h->nhandle);
		    free_shandle(h);
		} else
#endif
		    if (h->connection_time != 0 && !valid(h->player)) {
			log("RECYCLED: #%d on %s\n",
			    h->player,
			    network_connection_name(h->nhandle));
			network_send_line(h->nhandle, "*** Recycled ***");
			network_close(h->nhandle);
			free_shandle(h);
		    } else if (h->disconnect_me) {
			run_disfuncs(h->player);
			log("DISCONNECTED: %s on %s\n",
			    object_name(h->player),
			    network_connection_name(h->nhandle));
			network_send_line(h->nhandle, "*** Disconnected ***");
			network_close(h->nhandle);
			free_shandle(h);
		    }
	    }
	}
    }

    log("SHUTDOWN: %s\n", shutdown_message);
    send_shutdown_message(shutdown_message);
}

static shandle *
find_shandle(Objid player)
{
    shandle    *h;

    for (h = active_shandles; h; h = h->next)
	if (h->player == player)
	    return h;

    return 0;
}


/*
 * Exported interface
 */

static Objid 	next_unconnected_player = NOTHING - 1;

server_handle
server_new_connection(network_handle nh)
{
    shandle    	       *h = (shandle *) mymalloc(sizeof(shandle), M_NETWORK);
    server_handle	result;

    h->next = active_shandles;
    h->prev = &active_shandles;
    if (active_shandles)
	active_shandles->prev = &(h->next);
    active_shandles = h;

    h->nhandle = nh;
    h->connection_time = 0;
    h->last_activity_time = time(0);
    h->player = next_unconnected_player--;
    h->tasks = new_task_queue(h->player);
    h->disconnect_me = 0;

    enqueue_input_task(h->tasks, "");

    log("ACCEPT: #%d on %s\n", h->player, network_connection_name(nh));

    result.ptr = h;
    return result;
}

void
server_receive_line(server_handle sh, const char *line)
{
    shandle    *h = (shandle *) sh.ptr;

    h->last_activity_time = time(0);
    enqueue_input_task(h->tasks, line);
}

void
server_close(server_handle sh)
{
    shandle    *h = (shandle *) sh.ptr;
    Objid	player = h->player;

    log("CLIENT DISCONNECTED: %s on %s\n",
	object_name(player),
	network_connection_name(h->nhandle));
    free_shandle(h);
    run_disfuncs(player);
}

void
player_connected(Objid old_id, Objid new_id, int is_newly_created)
{
    shandle    *existing_h = find_shandle(new_id);
    shandle    *new_h = find_shandle(old_id);

    if (!new_h)
	panic("Non-existent shandle connected");
    
    new_h->player = new_id;
    new_h->connection_time = time(0);
    
    if (existing_h) {
	/* network_connection_name is allowed to reuse the same string
	 * storage, so we have to copy one of them.
	 */
	char   *name1 = str_dup(network_connection_name(existing_h->nhandle));

	log("REDIRECTED: %s, was %s, now %s\n",
	    object_name(new_id),
	    name1,
	    network_connection_name(new_h->nhandle));
	free_str(name1);
	network_send_line(existing_h->nhandle,
			  "*** Redirecting connection to new port ***");
	network_send_line(new_h->nhandle,
			  "*** Redirecting old connection to this port ***");
	network_close(existing_h->nhandle);
	free_shandle(existing_h);
    } else {
	log("%s: %s on %s\n",
	    is_newly_created ? "CREATED" : "CONNECTED",
	    object_name(new_h->player),
	    network_connection_name(new_h->nhandle));
	network_send_line(new_h->nhandle,
			  is_newly_created ? "*** Created ***"
					   : "*** Connected ***");
	run_confuncs(new_h->player);
    }
}

void
notify(Objid player, const char *message)
{
    shandle    *h = find_shandle(player);

    if (h)
	network_send_line(h->nhandle, message);
}

void
boot_player(Objid player)
{
    shandle    *h = find_shandle(player);

    if (h)
	h->disconnect_me = 1;
}

int
main(int argc, char **argv)
{
    if (argc < 3  ||  !network_initialize(argc - 3, argv + 3)) {
	errlog("Usage: %s input-db dump-db %s\n",
	       argv[0],
	       network_usage_string());
	exit(1);
    }

    log("STARTING: Version %s of the LambdaMOO server\n", server_version);
    log("          (Using %s protocol)\n", network_protocol_name());
    log("          (Task timeouts measured in %s seconds.)\n",
	virtual_timer_available() ? "server CPU" : "wall-clock");

    if (db_init(argv[1], argv[2]))
	exit(1);
    
    SRANDOM(time(0));

    parent_pid = getpid();
    setup_signals();
    reset_command_history();
    build_preptable();
    
    main_loop();

    network_shutdown();
    db_shutdown();

    return 0;
}


/**** built in functions ****/

static package
bf_server_version(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    r.type = TYPE_STR;
    r.v.str = str_dup(server_version);
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_renumber(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    Objid o = arglist.v.list[1].v.obj;
    free_var(arglist);
    
    if (!valid(o)) 
	return make_error_pack(E_INVARG);
    else if (!is_wizard(progr)) 
	return make_error_pack(E_PERM);
    
    r.type = TYPE_OBJ;
    r.v.obj = renumber_object(o);
    return make_var_pack(r);
}

static package
bf_reset_max_object(Var arglist, Byte next, void *vdata, Objid progr)
{
    free_var(arglist);
    
    if (!is_wizard(progr))
	return make_error_pack(E_PERM);
    
    reset_db_size();
    return no_var_pack();
}

static package
bf_memory_usage(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    r = memory_usage();
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_shutdown(Var arglist, Byte next, void *vdata, Objid progr)
{
    /*
     * The stream 's' and its contents will leak, but we're shutting down,
     * so it doesn't really matter.
     */

    Stream     *s = 0;

    if (!is_wizard(progr)) { 
	free_var(arglist);
	return make_error_pack(E_PERM);
    }
    
    s = new_stream(100);
    stream_printf(s, "shutdown() called by %s: %s",
		  object_name(progr),
		  arglist.v.list[1].v.str);
    shutdown_message = stream_contents(s);
    
    free_var(arglist);
    return no_var_pack();
}

static package
bf_dump_database(Var arglist, Byte next, void *vdata, Objid progr)
{
    free_var(arglist);
    if (!is_wizard(progr))
	return make_error_pack(E_PERM);
    
    checkpoint_requested = 1;
    return no_var_pack();
}

static package
bf_open_network_connection(Var arglist, Byte next, void *vdata, Objid progr)
{
#ifdef OUTBOUND_NETWORK

    Var		r;
    enum error	e;

    if (!is_wizard(progr))
	return make_error_pack(E_PERM);

    e = network_open_connection(arglist);
    if (e == E_NONE) {
	/* The connection was successfully opened, implying that
	 * server_new_connection was called, implying and a new negative
	 * player number was allocated for the connection.  Thus, the old
	 * value of next_unconnected_player is the number of our connection.
	 */
	r.type = TYPE_OBJ;
	r.v.obj = next_unconnected_player + 1;
    } else {
	r.type = TYPE_ERR;
	r.v.err = e;
    }
    if (r.type == TYPE_ERR)
	return make_error_pack(r.v.err);
    else
	return make_var_pack(r);

#else /* !OUTBOUND_NETWORK */

    /* This function is disabled in this server. */

    free_var(arglist);
    return make_error_pack(E_PERM);

#endif    
}

static package
bf_connected_players(Var arglist, Byte next, void *vdata, Objid progr)
{ /* () */
    shandle    *h;
    int		count = 0;
    Var		result;

    free_var(arglist);
    for (h = active_shandles; h; h = h->next)
	if (h->connection_time != 0)
	    count++;

    result = new_list(count);
    count = 0;

    for (h = active_shandles; h; h = h->next) {
	if (h->connection_time != 0) {
	    count++;
	    result.v.list[count].type = TYPE_OBJ;
	    result.v.list[count].v.obj = h->player;
	}
    }

    return make_var_pack(result);
}

static package
bf_connected_seconds(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (player) */
    Var r;
    shandle    *h = find_shandle(arglist.v.list[1].v.obj);
    
    r.type = TYPE_NUM;
    if (h  &&  h->connection_time != 0)
	r.v.num = time(0) - h->connection_time;
    else
	r.v.num = -1;
    free_var(arglist);
    if (r.v.num < 0)
	return make_error_pack(E_INVARG);
    else 
	return make_var_pack(r);
}

static package
bf_idle_seconds(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (player) */
    Var r;
    shandle    *h = find_shandle(arglist.v.list[1].v.obj);
    
    r.type = TYPE_NUM;
    if (h)
	r.v.num = time(0) - h->last_activity_time;
    else
	r.v.num = -1;
    free_var(arglist);
    if (r.v.num < 0)
	return make_error_pack(E_INVARG);
    else 
	return make_var_pack(r);
}

static package
bf_connection_name(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (player) */
    Objid	who = arglist.v.list[1].v.obj;
    shandle    *h = find_shandle(who);
    const char *conn_name;
    Var		r;

    if (h)
	conn_name = network_connection_name(h->nhandle);
    else
	conn_name = 0;
    
    free_var(arglist);
    if (!is_wizard(progr)  &&  progr != who)
	return make_error_pack(E_PERM);
    else if (!conn_name)
	return make_error_pack(E_INVARG);
    else {
	r.type = TYPE_STR;
	r.v.str = str_dup(conn_name);
	return make_var_pack(r);
    }
}

static package
bf_notify(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (player, string) */
    
    if (!is_wizard(progr) && progr != arglist.v.list[1].v.obj) {
	free_var(arglist);
	return make_error_pack(E_PERM);
    } 
    
    notify(arglist.v.list[1].v.obj, arglist.v.list[2].v.str);
    free_var(arglist);
    return no_var_pack();
}

static package
bf_boot_player(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object) */
    Objid oid = arglist.v.list[1].v.obj;
    
    free_var(arglist);
    
    if (oid != progr  &&  !is_wizard(progr)) 
	return make_error_pack(E_PERM);
    
    boot_player(oid);
    return no_var_pack();
}

void
register_server(void)
{ 
    (void) register_function("server_version", 0, 0, bf_server_version);
    (void) register_function("renumber", 1, 1, bf_renumber, TYPE_OBJ);
    (void) register_function("reset_max_object", 0, 0, bf_reset_max_object);
    (void) register_function("memory_usage", 0, 0, bf_memory_usage);
    (void) register_function("shutdown", 1, 1, bf_shutdown, TYPE_STR);
    (void) register_function("dump_database", 0, 0, bf_dump_database);
    (void) register_function("open_network_connection", 0, -1,
		      bf_open_network_connection);
    (void) register_function("connected_players", 0, 0, bf_connected_players);
    (void) register_function("connected_seconds", 1, 1, bf_connected_seconds,
			     TYPE_OBJ);
    (void) register_function("idle_seconds", 1, 1, bf_idle_seconds, TYPE_OBJ);
    (void) register_function("connection_name", 1, 1, bf_connection_name,
			     TYPE_OBJ);
    (void) register_function("notify", 2, 2, bf_notify, TYPE_OBJ, TYPE_STR);
    (void) register_function("boot_player", 1, 1, bf_boot_player, TYPE_OBJ);
}

char rcsid_server[] = "$Id: server.c,v 1.19 1992/10/23 23:03:47 pavel Exp $";

/* $Log: server.c,v $
 * Revision 1.19  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.18  1992/10/23  22:22:24  pavel
 * Change to #include "my-wait.h" instead of <sys/wait.h>.
 *
 * Revision 1.17  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.16  1992/10/17  20:52:15  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.15  1992/10/02  00:01:28  pavel
 * Fixed a bug in the main loop whereby external shutdown requests (e.g., the
 * interrupt signal) might not be noticed until some user I/O or a task execution occurred.
 *
 * Revision 1.14  1992/09/26  02:26:20  pavel
 * Added support for printing the network protocol name on server start-up.
 *
 * Revision 1.13  1992/09/15  15:53:00  pjames
 * Renamed FLUSH choices.
 *
 * Revision 1.12  1992/09/14  17:46:16  pjames
 * Moved db_modification code to db modules.
 * Added set up for disk based data base updating.
 *
 * Revision 1.11  1992/09/08  22:08:56  pjames
 * Added all code from bf_server.c.
 *
 * Revision 1.10  1992/09/03  23:52:33  pavel
 * Added support for multiple complete network implementations.
 *
 * Revision 1.9  1992/08/28  16:01:04  pjames
 * Changed myfree(*, M_STRING) to free_str(*).
 *
 * Revision 1.8  1992/08/26  18:30:45  pavel
 * Dump interval and retry interval are now local integers.
 * Changed logic in regards to using POSIX.
 *
 * Revision 1.7  1992/08/13  21:28:11  pjames
 * Now only calls one procedure to register built in functions.
 *
 * Revision 1.6  1992/08/12  01:50:42  pjames
 * run_ready_tasks() was moved to before logoff actions, so that logoffs
 * would happen faster.
 *
 * Revision 1.5  1992/08/10  16:55:31  pjames
 * Updated #includes.  Added register_bf_log().
 *
 * Revision 1.4  1992/07/30  23:22:26  pavel
 * Added a remote checkpoint-request signal (SIGUSR2), analogous to the existing
 * shutdown-requestion signal (SIGUSR1).
 *
 * Revision 1.3  1992/07/30  00:40:59  pavel
 * Add support for compiling on RISC/os 4.52 and NonStop-UX A22.
 *
 * Revision 1.2  1992/07/21  00:06:27  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
